#include <stdio.h>
#include <malloc.h>
#include <mpi.h>

#ifndef LIBELC

#define LIBELC

#define STRUCT_NAME_20 20
#define TAG_8 8

#define YES 1
#define NO 0

#define LOG_DONE 0
#define RECOVER_ON 1
#define CKPT_DONE 0
#define ENV_YES 1
#define ENV_DONE 2

#define CKPT_TAG 7777

//enum nType {INT,SHORT,LONG,DOUBLE,FLOAT,CHAR,STRUCTS,POINTER};
enum Property {NULLS,VAR,HEAP,PTR,FUNC};
enum JumpType {CALL,CKPT};

typedef char STRUCT[20];

struct VAR {
  int size;
  int count;
  void *pTop;
  void *pBottom;

  struct VAR *next;
  struct VAR *pre;
};

struct STRUCTS {
  STRUCT name;
  char *pAlias;  

  int ptrCount;
  int *ptrDisp;
  int size;

  struct STRUCTS *next;
  struct STRUCTS *pre;
};

struct POINTER
{
	void **pTarget;
	long offset;

	enum Property prop;
	int seq;
	
	struct POINTER *next;
	struct POINTER *pre;
};

struct FUNC 
{
	int seq;
	void *func;
//	struct VALUE *valHead;

	struct FUNC *next;
	struct FUNC *pre;
};

struct HEAP
{
   void *pTop;
   void *pBottom;
   int size;					//size
   
   struct POINTER *ptrHead;
   struct POINTER *ptrTail;
   int ptrCount;
   int *ptrDisp;

   struct HEAP *next;
   struct HEAP *pre;
};

struct FLOW {
	int Tag;				//tag's name
	int fNum;
	int subCount;
	int varCount;
	int ptrCount;
	int heapCount;

	struct FLOW *next;
	struct FLOW *pre;
};

// =====================================================================================================

//event logging node structure
struct MSG
{
	int count;
	int tag;
	int typeValue;
	int commValue;
	void *buf;
	MPI_Datatype *pType;
	MPI_Comm *pComm;

	struct MSG *next;
	struct MSG *pre;
};

//structure for saving the in-transit msg
struct LOG
{
	int count;
	int source;
	int tag;
	int typeValue;
	int commValue;
	MPI_Comm *pComm;
	MPI_Datatype *pType;

	void *buf;

	struct LOG *next;
	struct LOG *pre;
};

//structure containing all send/recv log link-table
//g_listHead points to process 0
struct LIST
{
	int flag;
	int rank;
	int count[4];				//0--send event number; 1--recv event number
	struct MSG *msgHead[4],*msgTail[4];		//0--send event log head; 1--recv event log head;
											//2--late				  3--early
	int log_counter;				//late msg number
	struct LOG *logHead,*logTail;		//late msg log head

	struct LIST *next;
	struct LIST *pre;
};

//structure for user defined derived datatype
//g_typeHead->next points to the 1st derived datatype
struct DATATYPE
{
	int typeValue;
	MPI_Datatype *pType;
	int freeFlag;

	struct DATATYPE *next;
	struct DATATYPE *pre;
};

struct COMM
{
	int commValue;
	MPI_Comm *pComm;
	int freeFlag;

	struct COMM *next;
	struct COMM *pre;
};

struct GROUP
{
	MPI_Group *pGroup; 
	int freeFlag;

	struct GROUP *next;
	struct GROUP *pre;	
};

// =======================================================================================================
int g_CKPT			=	0;			// dunno why
int g_RECOVER_FLAG	=	NO;			//flag to indicate the state of recover
int g_CKPT_FLAG		=	NO;				//flag to indicate the state of checkpoint
int g_LOG_COUNT		=	0;
int g_ENV_COUNT		=	0;
int g_RECOVER_COUNT	=	0;
int g_SIZE			=	0;
int g_MYRANK		=	0;
int g_TEMP			=	0;

int g_MSG_SIZE	=	sizeof(struct MSG);
int g_LOG_SIZE	=	sizeof(struct LOG);
int g_LIST_SIZE	=	sizeof(struct LIST);
int g_TYPE_SIZE	=	sizeof(struct DATATYPE);
int g_COMM_SIZE	=	sizeof(struct COMM);
int g_GROUP_SIZE	=	sizeof(struct GROUP);

int g_VAR_SIZE		=	sizeof(struct VAR);
int g_HEAP_SIZE		=	sizeof(struct HEAP);
int g_FLOW_SIZE		=	sizeof(struct FLOW);
int g_FUNC_SIZE		=	sizeof(struct FUNC);
int g_POINTER_SIZE	=	sizeof(struct POINTER);
int g_STRUCTS_SIZE	=	sizeof(struct STRUCTS);

int g_INT_SIZE		=	sizeof(int);
int g_DOUBLE_SIZE	=	sizeof(double);
int g_LONG_SIZE		=	sizeof(long);
int g_CHAR_SIZE		=	sizeof(char);
int g_SHORT_SIZE	=	sizeof(short);
int g_FLOAT_SIZE	=	sizeof(float);
int g_PTR_SIZE		=	sizeof(void *);
int g_BYTE_SIZE		=	1;

char g_f_VarName[20];
char g_f_HeapName[20];
char g_f_PtrName[20];
char g_f_FlowName[20];
char g_f_LogName[20];
char g_f_MsgName[20];
char g_f_MpiName[20];
char g_TYPE_PTR_NAME[1000]; 
char g_f_START='#';

struct LIST			*g_listHead,*g_listTail;
struct DATATYPE		*g_typeHead,*g_typeTail;
struct COMM			*g_commHead,*g_commTail;
struct GROUP		*g_groupHead,*g_groupTail;

struct FUNC			*g_funcHead,*g_funcTail;
struct HEAP			*g_heapHead,*g_heapTail;
struct POINTER		*g_ptrHead,*g_ptrTail;
struct VAR			*g_varHead,*g_varTail;
struct STRUCTS		*g_struHead,*g_struTail;
struct FLOW			*g_flowHead,*g_flowTail;

MPI_Datatype g_MPI_CHAR				=	MPI_CHAR;
MPI_Datatype g_MPI_SHORT			=	MPI_SHORT;
MPI_Datatype g_MPI_INT				=	MPI_INT;
MPI_Datatype g_MPI_LONG				=	MPI_LONG;
MPI_Datatype g_MPI_UNSIGNED_CHAR	=	MPI_UNSIGNED_CHAR;
MPI_Datatype g_MPI_UNSIGNED_SHORT	=	MPI_UNSIGNED_SHORT;
MPI_Datatype g_MPI_UNSIGNED			=	MPI_UNSIGNED;
MPI_Datatype g_MPI_UNSIGNED_LONG	=	MPI_UNSIGNED_LONG;
MPI_Datatype g_MPI_FLOAT			=	MPI_FLOAT;
MPI_Datatype g_MPI_DOUBLE			=	MPI_DOUBLE;
MPI_Datatype g_MPI_LONG_DOUBLE		=	MPI_LONG_DOUBLE;
MPI_Datatype g_MPI_BYTE				=	MPI_BYTE;
MPI_Datatype g_MPI_PACKED			=	MPI_PACKED;

MPI_Datatype g_MPI_FLOAT_INT		=	MPI_FLOAT_INT;
MPI_Datatype g_MPI_DOUBLE_INT		=	MPI_DOUBLE_INT;
MPI_Datatype g_MPI_LONG_INT			=	MPI_LONG_INT;
MPI_Datatype g_MPI_2INT				=	MPI_2INT;
MPI_Datatype g_MPI_SHORT_INT		=	MPI_SHORT_INT;
MPI_Datatype g_MPI_LONG_DOUBLE_INT	=	MPI_LONG_DOUBLE_INT;
MPI_Datatype g_MPI_LONG_LONG_INT	=	MPI_LONG_LONG_INT;

MPI_Datatype MPI_MSG;

MPI_Comm CKPT_COMM					=	MPI_COMM_WORLD;		
MPI_Comm g_MPI_COMM_WORLD;

#endif