#include <mpi.h>
//#include "libELC.h"
#include "data.h"
#include "var.c"
#include "heap.c"
#include "msg.c"
#include "libebc.c"

//------------------------------------------------------------------------------------------------------------

int ELC_MPI_Send(void *buf,int count,MPI_Datatype datatype,int target,int tag,MPI_Comm comm,int Tag,int fNum)
{
	int ierr;
	struct LIST *t_listTemp;

	OnCallEnter(Tag,fNum);
	if (g_CKPT_FLAG==YES) 
	{
		probe_others();	//if CKPT is ongoing, we try to catch CKPT request from other processes
	}
	t_listTemp=seek_list(target);
	insert_MSG(t_listTemp,count,datatype,tag,comm,0,buf);		//create the send event log
	ierr=MPI_Send(buf,count,datatype,target,tag,comm);

	return ierr;
}

int ELC_MPI_Isend(void* buf,int count,MPI_Datatype datatype,
				  int dest,int tag,MPI_Comm comm,MPI_request *request,
				  int Tag, int fNum)
{
	int ierr;
	struct LIST *t_listTempl

	if (g_CKPT_FLAG==YES) probe_others();

	t_listTemp=seek_list(target);
	insert_MSG(t_listTemp,count,datatype,tag,comm,0,buf);
	ierr=MPI_Isend(buf,count,datatype,dest,tag,comm,request);
}


int ELC_MPI_Recv(void *buf,int count,MPI_Datatype datatype,int source,int tag,MPI_Comm comm,MPI_Status *pStatus,
				 int Tag, int fNum)
{
	int ierr;
	struct LIST *t_listTemp;

	// ---------------------------------------RECOVER-----------------------------------------------------
	if (g_RECOVER_FLAG==YES)
	{
		if (source==MPI_ANY_SOURCE)
		{
			t_listTemp=g_listHead->next;
			while (t_listTemp!=NULL)
			{
				if (t_listTemp->log_counter>0) {
					if (try_log(t_listTemp,count,datatype,tag,comm,buf)) return;
				}
				if (t_listTemp->count[3]>0) {
					try_early(t_listTemp,count,datatype,source,tag,comm);
					break;
				}
				t_listTemp=t_listTemp->next;
			}
		}
		else {
				if (t_listTemp->log_counter>0) {
					if (try_log(t_listTemp,count,datatype,tag,comm,buf)) return;
				}
				if (t_listTemp->count[3]>0) try_early(t_listTemp,count,datatype,source,tag,comm);
		}

		if (t_listTemp->log_counter==0 && t_listTemp->count[3]==0) 
		{
			if (++g_RECOVER_COUNT==g_SIZE-1) g_RECOVER_FLAG=NO;
		}
	}

	// ---------------------------------------Receive------------------------------------------------------
	ierr=MPI_Recv(buf,count,datatype,source,tag,comm,pStatus);
	if (source==MPI_ANY_SOURCE || tag==MPI_ANY_TAG)
	{
		source=pStatus->MPI_SOURCE;
		tag=pStatus->MPI_TAG;
	}
	t_listTemp=seek_list(source);

	// -----------------------------------------CKPT--------------------------------------------------------
	if (g_CKPT_FLAG==YES)
	{
		probe_others();
		switch (t_listTemp->flag)
		{
			case (LOG_DONE): break;
			case (ENV_DONE): {
				if (seek_late_msg(count,datatype,tag,comm,t_listTemp)) 
					log_msg(buf,count,datatype,source,tag,comm,t_listTemp);
				break;
			}
			case (ENV_YES): log_msg(buf,count,datatype,source,tag,comm,t_listTemp);
		}
	}

	insert_MSG(t_listTemp,count,datatype,tag,comm,1,buf);

	return ierr;
}

int ELC_MPI_Init(int *argc,char ***argv)
{
	int count[4]={1,1,1,1};
	MPI_Aint size[4]={0,g_INT_SIZE,2*g_INT_SIZE,3*g_INT_SIZE};
	MPI_Datatype type[4]={MPI_INT,MPI_INT,MPI_INT,MPI_INT};

	int ierr=MPI_Init(argc,argv);

	ierr=MPI_Comm_size(MPI_COMM_WORLD,&g_SIZE);
	ierr=MPI_Comm_rank(MPI_COMM_WORLD,&g_MYRANK);
	MPI_Type_struct(4,count,size,type,&MPI_MSG);		//construct MPI_MSG
	MPI_Type_commit(&MPI_MSG);
	ierr=MPI_Comm_rank(MPI_COMM_WORLD,&g_MYRANK);

	sprintf(g_f_VarName,"./ckpt/var%d.dat",g_MYRANK);
	sprintf(g_f_HeapName,"./ckpt/heap%d.dat",g_MYRANK);
	sprintf(g_f_PtrName,"./ckpt/pointer%d.dat",g_MYRANK);
	sprintf(g_f_FlowName,"./ckpt/flow%d.dat",g_MYRANK);
	sprintf(g_f_LogName,"./ckpt/log%d.dat",g_MYRANK);
	sprintf(g_f_MsgName,"./ckpt/msg%d.dat",g_MYRANK);
	sprintf(g_f_MpiName,"./ckpt/mpi%d.dat",g_MYRANK);
	
	setup_LIST();							//setup the lists of event log and msg log
	if (g_RECOVER_FLAG==YES) RecoverFlow();
	
	return ierr;
}

int ELC_MPI_Finalize()
{
	MPI_Type_free(&MPI_MSG);
	return MPI_Finalize();
}

int ELC_MPI_Type_commit(MPI_Datatype *pDatatype)
{
	int ierr;
	struct DATATYPE *t_typeTemp=(struct DATATYPE *)malloc(g_TYPE_SIZE);

	ierr=MPI_Type_commit(pDatatype);
	if (ierr==0)	// commit succeeds
	{
		t_typeTemp->pType=pDatatype;
		t_typeTemp->typeValue=g_typeTail->typeValue+1;
		t_typeTemp->freeFlag=NO;
		t_typeTemp->pre=g_typeTail;
		t_typeTemp->next=NULL;
		
		g_typeTail=t_typeTemp;
	}

	return ierr;
}

int ELC_MPI_Type_free(MPI_Datatype *pDatatype)
{
	struct DATATYPE *t_typeTemp=g_typeTail;

	while (t_typeTemp!=g_typeHead)
	{
		if (t_typeTemp->pType==pDatatype)
		{
			if (g_CKPT_FLAG=YES)
			{	
				t_typeTemp->freeFlag=YES;
				pDatatype=NULL;
				return (0);							// postpone the actual type destruction
			}
			t_typeTemp->pre->next=t_typeTemp->next;
			if (t_typeTemp->next!=NULL) t_typeTemp->next->pre=t_typeTemp->pre;

			free(t_typeTemp);
			break;
		}
		t_typeTemp=t_typeTemp->pre;
	}

	return MPI_Type_free(pDatatype);
}

int ELC_MPI_Comm_dup(MPI_Comm comm,MPI_Comm *pComm)
{
	int ierr,myrank;
	struct COMM *t_commTemp=(struct COMM *)malloc(g_COMM_SIZE);
	MPI_Group t_groupTemp;

	MPI_Comm_group(comm,&t_groupTemp);
	MPI_Group_rank(t_groupTemp,&myrank);
	g_commHead->commValue++;

	if (myrank!=MPI_UNDEFINED) return;

	ierr=MPI_Comm_dup(comm,pComm);
	if (ierr==0)	// commit succeeds
	{
		t_commTemp->pComm=pComm;
		t_commTemp->freeFlag=NO;
		t_commTemp->commValue=g_commHead->commValue;

		t_commTemp->pre=g_commTail;
		t_commTemp->next=NULL;
		
		g_commTail=t_commTemp;
	}
	
	return ierr;
}

