

#include <stdio.h>
#include <malloc.h>
#include <mpi.h>

#ifndef LIBELC

#define LIBELC

#define STRUCT_NAME_20 20
#define TAG_8 8

#define YES 1
#define NO 0

#define LOG_DONE 0
#define RECOVER_ON 1
#define CKPT_DONE 0
#define ENV_YES 1
#define ENV_DONE 2

#define CKPT_TAG 7777

//enum nType {INT,SHORT,LONG,DOUBLE,FLOAT,CHAR,STRUCTS,POINTER};
enum Property {NULLS,VAR,HEAP,PTR,FUNC};
enum JumpType {CALL,CKPT};

typedef char STRUCT[20];

struct VAR {
  int size;
  int count;
  void *pTop;
  void *pBottom;

  struct VAR *next;
  struct VAR *pre;
};

struct STRUCTS {
  STRUCT name;
  char *pAlias;  

  int ptrCount;
  int *ptrDisp;
  int size;

  struct STRUCTS *next;
  struct STRUCTS *pre;
};

struct POINTER
{
	void **pTarget;
	long offset;

	enum Property prop;
	int seq;
	
	struct POINTER *next;
	struct POINTER *pre;
};

struct FUNC 
{
	int seq;
	void *func;
//	struct VALUE *valHead;

	struct FUNC *next;
	struct FUNC *pre;
};

struct HEAP
{
   void *pTop;
   void *pBottom;
   int size;					//size
   
   struct POINTER *ptrHead;
   int ptrCount;
   int *ptrDisp;

   struct HEAP *next;
   struct HEAP *pre;
};

struct FLOW {
	int Tag;				//tag's name
	int fNum;
	int subCount;
	int varCount;
	int ptrCount;
	int heapCount;

	struct FLOW *next;
	struct FLOW *pre;
};

// =====================================================================================================

//event logging node structure
struct MSG
{
	int count;
	int tag;
	int typeValue;
	int commValue;
	void *buf;
	MPI_Datatype *pType;
	MPI_Comm *pComm;

	struct MSG *next;
	struct MSG *pre;
};

//structure for saving the in-transit msg
struct LOG
{
	int count;
	int source;
	int tag;
	int typeValue;
	int commValue;
	MPI_Comm *pComm;
	MPI_Datatype *pType;

	void *buf;

	struct LOG *next;
	struct LOG *pre;
};

//structure containing all send/recv log link-table
//g_listHead points to process 0
struct LIST
{
	int flag;
	int rank;
	int count[4];				//0--send event number; 1--recv event number
	struct MSG *msgHead[4],*msgTail[4];		//0--send event log head; 1--recv event log head;
											//2--late				  3--early
	int log_counter;				//late msg number
	struct LOG *logHead,*logTail;		//late msg log head

	struct LIST *next;
	struct LIST *pre;
};

//structure for user defined derived datatype
//g_typeHead->next points to the 1st derived datatype
struct DATATYPE
{
	int typeValue;
	MPI_Datatype *pType;
	int freeFlag;

	struct DATATYPE *next;
	struct DATATYPE *pre;
};

struct COMM
{
	int commValue;
	MPI_Comm *pComm;
	int freeFlag;

	struct COMM *next;
	struct COMM *pre;
};

struct GROUP
{
	MPI_Group *pGroup; 
	int freeFlag;

	struct GROUP *next;
	struct GROUP *pre;	
};

// =======================================================================================================
extern int g_CKPT			;			// dunno why
extern int g_RECOVER_FLAG	;			//flag to indicate the state of recover
extern int g_CKPT_FLAG		;				//flag to indicate the state of checkpoint
extern int g_LOG_COUNT		;
extern int g_ENV_COUNT		;
extern int g_RECOVER_COUNT	;
extern int g_SIZE			;
extern int g_MYRANK			;
extern int g_TEMP			;

extern int g_MSG_SIZE		;
extern int g_LOG_SIZE		;
extern int g_LIST_SIZE		;
extern int g_TYPE_SIZE		;
extern int g_COMM_SIZE		;
extern int g_GROUP_SIZE		;

extern int g_VAR_SIZE		;
extern int g_HEAP_SIZE		;
extern int g_FLOW_SIZE		;
extern int g_FUNC_SIZE		;
extern int g_POINTER_SIZE	;
extern int g_STRUCTS_SIZE	;

extern int g_INT_SIZE		;
extern int g_DOUBLE_SIZE	;
extern int g_LONG_SIZE		;
extern int g_CHAR_SIZE		;
extern int g_SHORT_SIZE		;
extern int g_FLOAT_SIZE		;
extern int g_PTR_SIZE		;
extern int g_BYTE_SIZE		;

extern char g_f_VarName[20];
extern char g_f_HeapName[20];
extern char g_f_PtrName[20];
extern char g_f_FlowName[20];
extern char g_f_LogName[20];
extern char g_f_MsgName[20];
extern char g_f_MpiName[20];
extern char g_TYPE_PTR_NAME[1000]; 
extern char g_f_START	;

extern struct LIST			*g_listHead,*g_listTail;
extern struct DATATYPE		*g_typeHead,*g_typeTail;
extern struct COMM			*g_commHead,*g_commTail;
extern struct GROUP		*g_groupHead,*g_groupTail;

extern struct FUNC			*g_funcHead,*g_funcTail;
extern struct HEAP			*g_heapHead,*g_heapTail;
extern struct POINTER		*g_ptrHead,*g_ptrTail;
extern struct VAR			*g_varHead,*g_varTail;
extern struct STRUCTS		*g_struHead,*g_struTail;
extern struct FLOW			*g_flowHead,*g_flowTail;

extern MPI_Datatype g_MPI_CHAR				;
extern MPI_Datatype g_MPI_SHORT				;
extern MPI_Datatype g_MPI_INT				;
extern MPI_Datatype g_MPI_LONG				;
extern MPI_Datatype g_MPI_UNSIGNED_CHAR		;
extern MPI_Datatype g_MPI_UNSIGNED_SHORT	;
extern MPI_Datatype g_MPI_UNSIGNED			;
extern MPI_Datatype g_MPI_UNSIGNED_LONG		;
extern MPI_Datatype g_MPI_FLOAT				;
extern MPI_Datatype g_MPI_DOUBLE			;
extern MPI_Datatype g_MPI_LONG_DOUBLE		;
extern MPI_Datatype g_MPI_BYTE				;
extern MPI_Datatype g_MPI_PACKED			;

extern MPI_Datatype g_MPI_FLOAT_INT			;
extern MPI_Datatype g_MPI_DOUBLE_INT		;
extern MPI_Datatype g_MPI_LONG_INT			;
extern MPI_Datatype g_MPI_2INT				;
extern MPI_Datatype g_MPI_SHORT_INT			;
extern MPI_Datatype g_MPI_LONG_DOUBLE_INT	;
extern MPI_Datatype g_MPI_LONG_LONG_INT		;

extern MPI_Datatype MPI_MSG;

extern MPI_Comm CKPT_COMM;		
extern MPI_Comm g_MPI_COMM_WORLD;

#endif

void OnVarDef(int size, void *pTop, int count);
//-----------------------------------------------------------------------------------------------------------
// for structs
void OnStructDef(STRUCT name,int size,int ptrCount,int ptrDisp[],int strCount, int strDisp[],STRUCT strName[]);
void OnStructVarDef(STRUCT name,void *pTop,int count);

//-------------------------------------------------------------------------------------------------------------
// for pointer
void OnPtrDef(void *ptr,int count);

//--------------------------------------------------------------------------------------------------------------
// for function pointer
void OnFuncDef(void *func);

//--------------------------------------------------------------------------------------------------
//for typedef
void OnTypePtrDef(char *pTypePtrName);
void OnTypeOtherDef(char pOldName[STRUCT_NAME_20],char pNewName[STRUCT_NAME_20]);
void OnTypeVarDef(char pTypeName[STRUCT_NAME_20],int size,void *pTop,int count);

void *ELC_malloc(int size);
void *ELC_calloc(int nmemb, int size);
void *ELC_realloc(void *ptr, int size);
void ELC_free(void *ptr);

//-----------------------------------------------------------------------------------------------------
// for struct defined in the heap allocation
//-----------------------------------------------------------------------------------------------------
void OnHeapDef(STRUCT name, void *pTop, int count);

//-----------------------------------------------------------------------------------------------------------
// for type defined in the heap allocation
//-----------------------------------------------------------------------------------------------------------
void OnTypeHeapDef(char pTypeName[STRUCT_NAME_20],int size,void *pTop,int count);

void OnCallEnter(int Tag,int fNum);
void OnCallReturn();

void CKPT_Init();

int ELC_MPI_Send(void *buf,int count,MPI_Datatype datatype,int target,int tag,MPI_Comm comm,int Tag,int fNum);
int ELC_MPI_Recv(void *buf,int count,MPI_Datatype datatype,int source,int tag,MPI_Comm comm,MPI_Status *pStatus,
				 int Tag, int fNum);

int ELC_MPI_Init(int *argc,char ***argv);
int ELC_MPI_Finalize();

int ELC_MPI_Type_commit(MPI_Datatype *pDatatype);
int ELC_MPI_Type_free(MPI_Datatype *pDatatype);

int ELC_MPI_Comm_dup(MPI_Comm comm,MPI_Comm *pComm);